# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
#
#   python-signalk-client is a python client library for SignalK
#   Copyright (C) 2016  Philip J Freeman <elektron@halo.nu>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from signalk_client.datum import Datum

"""vessel objects"""

class Vessel(object):
    """Vessel Object

    Keyword arguments:
    data -- SignalK Data object
    vessel_key -- dictionary key for vessel
    """
    def __init__(self, data, vessel_key):
        self.key = vessel_key
        self.data = data
        self._load_vessel_data()

    def __str__(self):
        if self.name:
            return self.name
        elif self.mmsi:
            return "MMSI: {}".format(self.mmsi)
        elif self.uuid:
            return "UUID: {}".format(self.uuid)
        elif self.url:
            return "URL: {}".format(self.url)

        return "KEY: {}".format(self.key)

    def _load_vessel_data(self):
        """load vessel data from signalk"""
        # set some defaults
        self.uuid = None
        self.mmsi = None
        self.url = None
        self.name = None

        # lookup stuff
        try:
            self.uuid = self.get_datum('uuid').display_value()
        except (KeyError, TypeError):
            pass

        try:
            self.mmsi = self.get_datum('mmsi').display_value()
        except (KeyError, TypeError):
            pass

        try:
            self.url = self.get_datum('url').display_value()
        except (KeyError, TypeError):
            pass

        #SignalK Requires one of the above settings
        assert(self.uuid or self.mmsi or self.url)

        try:
            self.name = self.get_datum('name').display_value()
        except (KeyError, TypeError):
            pass

    def get_targets(self):
        """returns a list of available properties for vessel
        """
        out_targets = []
        for path in self.data.meta.keys():

            if path.find("*") > -1:

                # TODO: Support wildcard paths in target scan
                logging.warning(
                    "Unhandled wildcard in target search for: {}".format(path)
                    )

            elif path.find("$") > -1:

                # TODO: Support reference paths in target scan ???
                logging.warning(
                    "Unhandled reference in target search for: {}".format(path)
                    )

            else:

                try:
                    self.get_datum(path)
                except KeyError:
                    continue

                out_targets.append(path)

        return sorted(out_targets)

    def get_prop(self, path):
        """get raw data for vessel's property

        Keyword arguments:
            path -- a signalk property path

        this returns a raw object (or whatever data exists) for the provided
        signalk property path.
        """
        return self.data.get_by_map_list(
            ['vessels', self.key] + path.split('.')
            )

    def get_datum(self, path):
        """get a Datum object for vessel's property
        """
        value = None
        units = None
        desc = None

        got = self.get_prop(path)

        if isinstance(got, dict):
            value = got['value']
        else:
            value = got

        if self.data.meta.has_key(path):
            if self.data.meta[path].has_key('units'):
                units = self.data.meta[path]['units']
            if self.data.meta[path].has_key('description'):
                desc = self.data.meta[path]['description']
        return Datum(path, value, units, desc)



# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
#
#   python-signalk-client is a python client library for SignalK
#   Copyright (C) 2016  Philip J Freeman <elektron@halo.nu>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""signalk client connection"""

import json
import logging
import requests
import threading
import websocket
from signalk_client.data import Data

class Client(object):
    """Client connection object

    Keyword arguments:
    server -- server host (default None)

    By default, this function uses `zeroconf_server` to automatically locate
    a _signalk-http._tcp.local. service on the local network.

    Optionally, you can pass the hostname of the signalk server via the
    keyword `server`. To specify a port use `hostname:port` format.
    """

    def __init__(self, server=None):
        self.server = server
        self.api_endpoint = None
        self.stream_endpoint = None

        self.__config()

        self.data = Data(requests.get(self.api_endpoint).json())
        self.w_sock = websocket.WebSocketApp(
            "%s?subscribe=all"%(self.stream_endpoint),
            on_message=self.__ws_on_message,
            on_error=self.__ws_on_error,
            on_close=self.__ws_on_close,
            on_open=self.__ws_on_open
            )

        self.websocket_t = threading.Thread(target=self.w_sock.run_forever)
        self.websocket_t.daemon = True
        self.websocket_t.start()

    def __zeroconf_server(self):
        """discover local signalk server
        """

        import time
        from signalk_client.zeroconf import ServiceBrowser, Zeroconf

        class MyListener(object):
            """zeroconf listener object"""
            def __init__(self):
                self.services = {}

            def remove_service(self, zeroconf, service_type, name):
                """service removal function"""
                self.services.pop(name)
                logging.info("zeroconf: service removed: {}".format(name))

            def add_service(self, zeroconf, service_type, name):
                """service add function"""
                info = zeroconf.get_service_info(service_type, name)
                self.services[name] = info
                logging.info("zeroconf: service found: {} @ {}:{}".format(
                    name,
                    info.server,
                    info.port
                    ))

            def get_services(self):
                """service getter"""
                return self.services

        service_type = "_signalk-http._tcp.local."
        zeroconf = Zeroconf()
        listener = MyListener()
        browser = ServiceBrowser(zeroconf, service_type, listener)
        while True:
            time.sleep(2)
            if len(listener.get_services().keys()) > 0:
                break
            logging.warning("No Services of type:{} found.. waiting..".format(
                service_type
                ))
        zeroconf.close()
        services = listener.get_services()
        service_info = services[services.keys()[0]]
        return "%s:%s"%(service_info.server, service_info.port)

    def __config(self):
        """discover endpoints from server
        """

        import requests

        # Set from arg or discover via zeroconf, server address and port
        if self.server == None:
            self.server = self.__zeroconf_server()
        logging.info("Attempting Connection to {}...".format(self.server))

        # Discover API and Stream endpoints
        data = requests.get("http://%s/signalk"%(self.server)).json()
        endpoints = data['endpoints']['v1']
        logging.info("Connected to SignalK Server({})".format(
            endpoints['version']
            ))

        self.api_endpoint = endpoints['signalk-http']
        self.stream_endpoint = endpoints['signalk-ws']

        logging.info("Got endpoints: api_endpoint={} stream_endpoint={}".format(
            self.api_endpoint,
            self.stream_endpoint,
            ))

    def __ws_on_message(self, w_sock, message):
        """websocket message handler"""
        self.data.process_delta(json.loads(message))

    def __ws_on_error(self, w_sock, error):
        """websocket error handler"""
        logging.error("websocket error: {}".format(error))

    def __ws_on_close(self, w_sock):
        """websocket connection close handler"""
        # TODO: handle connection close by reconnecting?
        logging.warning("websocket closed")

    def __ws_on_open(self, w_sock):
        """websocket connection open handler"""
        pass

    def close(self):
        """close the signalk client connection
        """
        logging.warning("Closing websocket...")
        self.w_sock.close()



# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
#
#   python-signalk-client is a python client library for SignalK
#   Copyright (C) 2016  Philip J Freeman <elektron@halo.nu>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""dealing with individual data"""

import math
import re

FEET_PER_METER = 3.28084
METERS_PER_NAUTICAL_MILE = 1852
SECONDS_PER_HOUR = 60*60
KELVIN_CELSIUS_OFFSET = 273.15

DISPLAY_PATH_CONVERSIONS = [
    (re.compile(r"^environment\."), "env."),
    (re.compile(r"^navigation\."), "nav."),
    (re.compile(r"\.position\."), ".pos."),
    ]

def convert(value, from_unit, to_unit):
    """helper function to handle unit conversion"""

    if from_unit == 'm' and to_unit == 'ft':
        return value*FEET_PER_METER
    elif from_unit == 'm/s' and to_unit == 'kn':
        return value*SECONDS_PER_HOUR/METERS_PER_NAUTICAL_MILE
    elif from_unit == 'rad' and to_unit == 'deg':
        return 180 * value / math.pi
    elif from_unit == 'K' and to_unit == 'C':
        return value-KELVIN_CELSIUS_OFFSET
    elif from_unit == 'K' and to_unit == 'F':
        return ((value-KELVIN_CELSIUS_OFFSET)*(9.0/5.0))+32.0
    else:
        raise NotImplementedError(
            "Conversion from {} to {} is not supported".format(
                from_unit,
                to_unit
                )
            )

class Datum(object):
    """class for dealing with individual datums
    """
    def __init__(self, path, value=None, units=None, desc=None):
        self.path = path
        self.value = value
        self.desc = desc
        self.units = units

    def __str__(self):
        return "{}: {}".format(
            self.display_path(),
            self.display_value()
            )

    def display_path(self):
        """return the property path for display"""
        out_string = self.path
        for pattern, replacement in DISPLAY_PATH_CONVERSIONS:
            out_string = re.sub(pattern, replacement, out_string)
        return out_string

    def display_value(self, convert_units=None):
        """return the property value for display

        convert_units is a list of conversions to make

            convert_units = [('m', 'ft'), ('m/s', 'kn')]
        """
        value = self.value
        units = self.units

        if convert_units != None:
            for from_unit, to_unit in convert_units:
                if units == from_unit:
                    value = convert(value, from_unit, to_unit)
                    units = to_unit
        out_string = ""
        if value == None:
            out_string += "--"
        else:
            if self.path.endswith('.latitude'):
                direction = 'N' if value >= 0 else 'S'
                out_string += "{:.5f} {}".format(abs(value), direction)
            elif self.path.endswith('.longitude'):
                direction = 'E' if value >= 0 else 'W'
                out_string += "{:.5f} {}".format(abs(value), direction)
            elif isinstance(value, float):
                out_string += "{:.2f}".format(value)
            else:
                out_string += "{}".format(value)
            if units != None:
                out_string += " {}".format(units)
        return out_string



# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import unittest
import signalk_client.data as data

class TestDataInitialization(unittest.TestCase):

    def setUp(self):
        self.data = data.Data()

    def test_data_initial_structure(self):
        self.assertEqual(self.data.data.has_key('version'), True);
        self.assertEqual(type(self.data.data['version']), str);
        self.assertEqual(self.data.data.has_key('vessels'), True);
        self.assertEqual(type(self.data.data['vessels']), list);



# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import unittest
import signalk_client.datum as datum

class TestDatumConvert(unittest.TestCase):

    def test_datum_convert_meters_to_feet(self):
        self.assertEqual(datum.convert(1.0, 'm', 'ft'), 3.28084)

    def test_datum_convert_meters_per_second_to_knots(self):
        self.assertAlmostEqual(datum.convert(1.0, 'm/s', 'kn'), 1.94384, places=5)

    def test_datum_convert_radians_to_degrees(self):
        self.assertAlmostEqual(datum.convert(1.0, 'rad', 'deg'), 57.2958, places=4)

    def test_datum_convert_kelvin_to_celsius(self):
        self.assertAlmostEqual(datum.convert(1.0, 'K', 'C'), -272.15, places=2)

    def test_datum_convert_kelvin_to_fahrenheit(self):
        self.assertAlmostEqual(datum.convert(1.0, 'K', 'F'), -457.87, places=2)

    def test_datum_convert_raises_notimplementederror(self):
        with self.assertRaises(NotImplementedError):
            datum.convert(1.0, 'blah', 'bleh')



# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
#
#   python-signalk-client is a python client library for SignalK
#   Copyright (C) 2016  Philip J Freeman <elektron@halo.nu>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""signalk data"""

import copy
import json
import logging
import pkg_resources
from signalk_client.vessel import Vessel

class Data(object):
    """signalk data object
    """

    def __init__(self, seed=None):

        self.meta = json.load(pkg_resources.resource_stream(
            'signalk_client', 'include/meta.json'
            ))

        if seed == None:
            # bare minimum valid schema
            self.data = {'version': "0.1.0", 'vessels': []}

        else:
            self.data = seed

            logging.debug(
                "loaded initial data:\n{}".format(
                    json.dumps(self.data, indent=2, sort_keys=True)
                    )
                )


    def process_delta(self, data):
        """Parse SignalK Delta message and update data store
        """

        if 'self' in data and 'version' in data and 'timestamp' in data:
            # hello message
            logging.debug(
                "hello message in delta stream: {!r}".format(data)
                )
            context_list = []
            update_properties = {}
            if 'self' in data:
                logging.info("setting self = {!r}".format(data['self']))
                self.__set_by_map_list(['self'], data['self'])
        elif 'updates' in data:
            # delta message
            if 'context' in data:
                context_list = data['context'].split(".")
            else:
                context_list = []
            for update in data['updates']:
                update_properties = {}
                if 'source' in update:
                    update_properties['source'] = update['source']
                if '$source' in update:
                    update_properties['$source'] = update['$source']
                if 'timestamp' in update:
                    update_properties['timestamp'] = update['timestamp']
                for value in update['values']:
                    item_in = copy.deepcopy(update_properties)
                    map_list = copy.deepcopy(context_list)
                    map_list += value['path'].split(".")
                    if isinstance(value['value'], dict):
                        for key in value['value'].keys():
                            item_in[key] = value['value'][key]
                    else:
                        item_in['value'] = value['value']
                    self.__set_by_map_list(map_list, item_in)
            logging.debug("updated context: {}\n{}".format(
                ".".join(context_list),
                json.dumps(
                    self.get_by_map_list(context_list),
                    indent=4, sort_keys=True
                    )
                ))
        else:
            logging.warning(
                "ignoring unrecognized delta message: {!r}".format(data)
                )

    def get_by_map_list(self, map_list):
        """return a data object from a hierarchical list of keys"""
        value = reduce(lambda d, k: d[k], map_list, self.data)
        return value

    def __set_by_map_list(self, map_list, value):
        """set a data object at location based on a hierarchical list of keys"""
        self.get_by_map_list(map_list[:-1])[map_list[-1]] = value

    def get_prop_meta(self, path):
        """get meta-data object from a property path
        """
        if self.meta.has_key(path):
            return self.meta[path]
        else:
            return {}

    def get_vessels(self):
        """returns a list of vessels (as Vessel objects) signalk knows of
        """
        vessels = []
        for vessel_key in self.data['vessels'].keys():
            vessels.append(Vessel(self, vessel_key))
        return vessels

    def get_self(self):
        """returns "self" vessel (as Vessel object)
        """
        vessels = []
        for vessel_key in self.data['vessels'].keys():
            vessels.append(Vessel(self, vessel_key))

        return Vessel(self, self.data['self'])
